@@ -1,29 +1,26 @@
 # -*- coding: utf-8 -*-
 
-import json
-import grpc
 from concurrent import futures
+from typing import Any, List, Optional, Callable
 
-from pip_services3_commons.run.IOpenable import IOpenable
-from pip_services3_commons.config.IConfigurable import IConfigurable
-from pip_services3_commons.refer.IReferenceable import IReferenceable
-from pip_services3_commons.refer.IReferences import IReferences
+import grpc
 from pip_services3_commons.config.ConfigParams import ConfigParams
-from pip_services3_commons.run.Parameters import Parameters
-from pip_services3_components.log.CompositeLogger import CompositeLogger
-from pip_services3_components.count.CompositeCounters import CompositeCounters
-from pip_services3_commons.errors.ErrorDescriptionFactory import ErrorDescriptionFactory
+from pip_services3_commons.config.IConfigurable import IConfigurable
 from pip_services3_commons.errors.ConnectionException import ConnectionException
+from pip_services3_commons.errors.ErrorDescriptionFactory import ErrorDescriptionFactory
 from pip_services3_commons.errors.InvocationException import InvocationException
-from pip_services3_commons.convert.JsonConverter import JsonConverter
+from pip_services3_commons.refer import IReferences
+from pip_services3_commons.refer.IReferenceable import IReferenceable
+from pip_services3_commons.run.IOpenable import IOpenable
+from pip_services3_commons.run.Parameters import Parameters
+from pip_services3_commons.validate import Schema
+from pip_services3_components.count.CompositeCounters import CompositeCounters
+from pip_services3_components.log.CompositeLogger import CompositeLogger
 from pip_services3_rpc.connect.HttpConnectionResolver import HttpConnectionResolver
-from pip_services3_commons.validate.Schema import Schema
-from pip_services3_commons.data.DataPage import DataPage
-
-from .IRegisterable import IRegisterable
 
 import pip_services3_grpc.protos.commandable_pb2 as commandable_pb2
 import pip_services3_grpc.protos.commandable_pb2_grpc as commandable_pb2_grpc
+from pip_services3_grpc.services.IRegisterable import IRegisterable
 
 
 class _CommandableMediator(commandable_pb2_grpc.CommandableServicer):
@@ -33,7 +30,7 @@ class _CommandableMediator(commandable_pb2_grpc.CommandableServicer):
     def __init__(self):
         self.__invoke = None
 
-    def invoke(self, request, context):
+    def invoke(self, request: commandable_pb2.InvokeRequest, context: grpc.ServicerContext):
         if self.__invoke is not None:
             return self.__invoke(request, context)
 
@@ -72,8 +69,19 @@ class GrpcEndpoint(IOpenable, IConfigurable, IReferenceable):
         - counters: **"\*:counters:\*:\*:1.0"**;
         - discovery: **"\*:discovery:\*:\*:1.0"** (for the connection resolver).
 
-    ### Examples ###
-        TODO!!!
+    .. code-block:: python
+
+        def my_method(self, _config, _references):
+            endpoint = GrpcEndpoint()
+            if self._config:
+                endpoint.configure(self._config)
+            if self._references:
+                endpoint.set_references(self._references)
+            ...
+
+            self._endpoint.open(correlation_id)
+            ...
+
 
     """
 
@@ -90,23 +98,23 @@ class GrpcEndpoint(IOpenable, IConfigurable, IReferenceable):
         "options.request_max_size", 1024 * 1024,
         "options.file_max_size", 200 * 1024 * 1024,
         "options.connect_timeout", 60000,
-        "options.debug", None
+        "options.debug", True
     )
 
-    __server = None
-    __connection_resolver = HttpConnectionResolver()
-    __logger = CompositeLogger()
-    __counters = CompositeCounters()
-    __maintenance_enabled = False
-    __file_max_size = 200 * 1024 * 1024
-    __uri = None
-    __registrations = []
-    __commandable_methods = None
-    __commandable_schemas = None
-    __commandable_service = None
-    __interceptors = []
+    def __init__(self):
+        self.__server: Any = None
+        self.__connection_resolver = HttpConnectionResolver()
+        self.__logger = CompositeLogger()
+        self.__counters = CompositeCounters()
+        self.__maintenance_enabled = False
+        self.__file_max_size = 200 * 1024 * 1024
+        self.__uri: str = None
+        self.__registrations: List[IRegisterable] = []
+        self.__commandable_methods: Any = None
+        self.__commandable_schemas: Any = None
+        self.__commandable_service: Any = None
 
-    def configure(self, config):
+    def configure(self, config: ConfigParams):
         """
         Configures this HttpEndpoint using the given configuration parameters.
         
@@ -131,7 +139,7 @@ class GrpcEndpoint(IOpenable, IConfigurable, IReferenceable):
         self.__file_max_size = ConfigParams().get_as_long_with_default(key='options.file_max_size',
                                                                        default_value=self.__file_max_size)
 
-    def set_references(self, references):
+    def set_references(self, references: IReferences):
         """
         Sets references to this endpoint's logger, counters, and connection resolver.
 
@@ -146,31 +154,30 @@ class GrpcEndpoint(IOpenable, IConfigurable, IReferenceable):
         self.__counters.set_references(references)
         self.__connection_resolver.set_references(references)
 
-    def is_open(self):
+    def is_open(self) -> bool:
         """
         :return: whether or not this endpoint is open with an actively listening GRPC server.
         """
         return self.__server is not None
 
-    def open(self, correlation_id, max_workers=15):
+    def open(self, correlation_id: Optional[str]):
         """
         Opens a connection using the parameters resolved by the referenced connection
         resolver and creates a GRPC server (service) using the set options and parameters.
 
         :param correlation_id: (optional) transaction id to trace execution through call chain.
-        :param max_workers: max count of workers in thread pool.
         """
         if self.is_open():
             return
 
         connection = self.__connection_resolver.resolve(correlation_id)
-        self.__uri = connection.get_uri()
+        self.__uri = connection.get_as_string('uri')
         try:
             self.__connection_resolver.register(correlation_id)
 
             credentials = None
 
-            if connection.get_protocol_with_default('http') == 'https':
+            if connection.get_as_string_with_default('protocol', 'http') == 'https':
                 ssl_key_file = connection.get_as_nullable_string('ssl_key_file')
                 ssl_crt_file = connection.get_as_nullable_string('ssl_crt_file')
 
@@ -180,35 +187,45 @@ class GrpcEndpoint(IOpenable, IConfigurable, IReferenceable):
                 with open(ssl_crt_file, 'rb') as file:
                     certificate = file.read()
 
-                credentials = grpc.ssl_server_credentials(((private_key, certificate),))
+                ca = None
+                ssl_ca_file = connection.get_as_nullable_string('ssl_ca_file')
+                if ssl_ca_file is not None:
+                    with open(ssl_ca_file, 'rb') as file:
+                        ca_text = file.read()
+                    while ca_text and len(ca_text.strip()) > 0:
+                        crt_index = ca_text.rfind(b'-----BEGIN CERTIFICATE-----')
+                        if crt_index > -1:
+                            ca = ca_text[crt_index:]
+                            ca_text = ca_text[:crt_index]
+
+                credentials = grpc.ssl_server_credentials(((private_key, certificate),), root_certificates=ca)
 
             # Create instance of express application
-            if len(self.__interceptors) > 0:
-                self.__server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers),
-                                            interceptors=self.__interceptors)
-            else:
-                self.__server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
+            self.__server = grpc.server(futures.ThreadPoolExecutor())
 
             if credentials:
-                self.__server.add_secure_port(str(connection.get_host()) + ':' +
-                                              str(connection.get_port()), credentials)
+                self.__server.add_secure_port(str(connection.get_as_string('host')) + ':' +
+                                              str(connection.get_as_string('port')), credentials)
             else:
-                self.__server.add_insecure_port(str(connection.get_host()) + ':' + str(connection.get_port()))
-            # Start operations
-            self.__server.start()
+                self.__server.add_insecure_port(
+                    str(connection.get_as_string('host')) + ':' + str(connection.get_as_string('port')))
+
             self.__connection_resolver.register(correlation_id)
-            self.__perform_registrations()
             self.__logger.debug(correlation_id, 'Opened GRPC service at {}'.format(self.__uri))
 
+            # Start operations
+            self.__perform_registrations()
+            self.__server.start()
+
         except Exception as ex:
             self.__server = None
-            err = ConnectionException(correlation_id, 'CANNOT_CONNECT', 'Opening GRPC service failed').wrap(
+            raise ConnectionException(correlation_id, 'CANNOT_CONNECT', 'Opening GRPC service failed').wrap(
                 ex).with_details('url', self.__uri)
-            raise err
 
-    def close(self, correlation_id):
+    def close(self, correlation_id: Optional[str]):
         """
         Closes this endpoint and the GRPC server (service) that was opened earlier.
+
         :param correlation_id: (optional) transaction id to trace execution through call chain.
         """
         if self.__server is not None:
@@ -222,23 +239,21 @@ class GrpcEndpoint(IOpenable, IConfigurable, IReferenceable):
                 self.__server.stop(None)
                 self.__logger.debug(correlation_id, 'Closed GRPC service at {}'.format(self.__uri))
                 self.__server = None
-                GrpcEndpoint.__interceptors = []
-                GrpcEndpoint.__registrations = []
-                GrpcEndpoint.__connection_resolver = HttpConnectionResolver()
+                self.__registrations = []
+                self.__connection_resolver = HttpConnectionResolver()
             except Exception as ex:
                 self.__logger.warn(correlation_id, 'Failed while closing GRPC service: '.format(ex))
 
-    def register(self, registration):
+    def register(self, registration: IRegisterable):
         """
         Registers a registerable object for dynamic endpoint discovery.
 
         :param registration: the registration to add.
-        :return:
         """
         if registration is not None:
             self.__registrations.append(registration)
 
-    def unregister(self, registration):
+    def unregister(self, registration: IRegisterable):
         """
         Unregisters a registerable object, so that it is no longer used in dynamic
         endpoint discovery.
@@ -251,6 +266,10 @@ class GrpcEndpoint(IOpenable, IConfigurable, IReferenceable):
         for registration in self.__registrations:
             registration.register()
 
+            # hack for register generated services
+            if hasattr(registration, '_GrpcService__register_service') and hasattr(registration, 'add_servicer_to_server'):
+                registration._GrpcService__register_service(registration)
+
         self.__register_commandable_service()
 
     def __register_commandable_service(self):
@@ -261,9 +280,9 @@ class GrpcEndpoint(IOpenable, IConfigurable, IReferenceable):
 
         self.register_service(self.__commandable_service)
 
-    def __invoke_commandable_method(self, request, context):
+    def __invoke_commandable_method(self, request: commandable_pb2.InvokeRequest, context: grpc.ServicerContext):
         method = request.method
-        action = self.__commandable_methods[method] if self.__commandable_methods else None
+        action = None if not self.__commandable_methods else self.__commandable_methods[method]
         correlation_id = request.correlation_id
         response = commandable_pb2.InvokeReply()
 
@@ -271,82 +290,35 @@ class GrpcEndpoint(IOpenable, IConfigurable, IReferenceable):
         if action is None:
             err = InvocationException(correlation_id, 'METHOD_NOT_FOUND',
                                       'Method ' + method + ' was not found').with_details('method', method)
+
             resp_err = ErrorDescriptionFactory.create(err)
-            response.error = commandable_pb2.ErrorDescription()
-            response.error.category = resp_err.category
-            response.error.code = resp_err.code
-            response.error.correlation_id = resp_err.correlation_id
-            response.error.status = resp_err.status
-            response.error.message = resp_err.message
-            response.error.cause = resp_err.cause
-            response.error.stack_trace = resp_err.stack_trace
-            response.error.details.addAll(resp_err.details)
-            response.resultEmpty = True
-            response.resultJson = ''
-
-            return response
-
-        try:
-            # Convert arguments
-            args_empty = request.args_empty
-            args_json = request.args_json
-            args = Parameters.from_json(args_json) if not args_empty and args_json else Parameters()
-
-            try:
-                result = action(correlation_id, args, method)
-                response.result_empty = result is None
-                if type(result) is DataPage:
-                    response.result_json = json.dumps(result.to_json())
-                else:
-                    response.result_json = json.dumps(result) if result is not None else ''
-
-                # TODO: Validate schema
-                schema = self.__commandable_schemas[method]
-                if schema:
-                    pass
-
-            except Exception as ex:
-                # Process result and generate response
-                resp_err = ErrorDescriptionFactory.create(ex)
-                response.error.category = resp_err.category
-                response.error.code = resp_err.code
-                response.error.correlation_id = resp_err.correlation_id
-                response.error.status = resp_err.status
-                response.error.message = resp_err.message
-                response.error.cause = resp_err.cause
-                response.error.stack_trace = resp_err.stack_trace
-                response.error.details.update(resp_err.details)
-                response.result_empty = True
-                response.result_json = ''
-
-        except Exception as ex:
-            # Handle unexpected exception
-            err = InvocationException(correlation_id, 'METHOD_FAILED', 'Method ' + method + ' failed').wrap(
-                ex).with_details('method', method)
-            resp_err = ErrorDescriptionFactory.create(err)
-            response.error.category = resp_err.category
-            response.error.code = resp_err.code
-            response.error.correlation_id = resp_err.correlation_id
-            response.error.status = resp_err.status
-            response.error.message = resp_err.message
-            response.error.cause = resp_err.cause
-            response.error.stack_trace = resp_err.stack_trace
-            response.error.details.update(resp_err.details)
+            response.error.category = resp_err.category or ''
+            response.error.code = resp_err.code or ''
+            response.error.correlation_id = resp_err.correlation_id or ''
+            response.error.status = resp_err.status or 0
+            response.error.message = resp_err.message or ''
+            response.error.cause = resp_err.cause or ''
+            response.error.stack_trace = resp_err.stack_trace or ''
+            response.error.details.update(resp_err.details or {})
             response.result_empty = True
             response.result_json = ''
 
-        return response
+            return response
 
-    def register_service(self, service):
+        return action(request, context)
+
+    def register_service(self, service: Any):
         """
         Registers a service with related implementation
+
         :param service: a GRPC service object.
         """
         service.add_servicer_to_server(self.__server)
 
-    def register_commandable_method(self, method, schema, action):
+    def _register_commandable_method(self, method: str, schema: Schema,
+                                     action: Callable[[Optional[str], Any], Any]):
         """
-        Registers a commandable method in this objects GRPC server (service) by the given name.,
+        Registers a commandable method in this objects GRPC server (service) by the given name.
 
         :param method: the GRPC method name.
         :param schema: the schema to use for parameter validation.
@@ -357,6 +329,3 @@ class GrpcEndpoint(IOpenable, IConfigurable, IReferenceable):
 
         self.__commandable_schemas = self.__commandable_schemas or {}
         self.__commandable_schemas[method] = schema
-
-    def register_interceptor(self, interceptor):
-        self.__interceptors.append(interceptor)
